@@ -1,4 +1,4 @@
-# (C) Copyright 2018-2022 by Rocky Bernstein
+# (C) Copyright 2018-2024 by Rocky Bernstein
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU General Public License
@@ -37,9 +37,10 @@ import struct
 import sys
 from typing import Dict
 
-from xdis.version_info import IS_PYPY, version_tuple_to_str
+from xdis.version_info import IS_GRAAL, IS_PYPY, version_tuple_to_str
 
-IS_PYPY3 = (48, 64, 112, 160, 192, 240, 244, 256, 336)
+PYPY3_MAGICS = (48, 64, 112, 160, 192, 240, 244, 256, 336, 384)
+GRAAL3_MAGICS = (21150, 21280)
 
 
 def add_magic_from_int(magic_int, version):
@@ -77,13 +78,13 @@ def magic2int(magic: bytes) -> int:
     return struct.unpack("<Hcc", magic)[0]
 
 
-def __by_version(magics) -> dict:
-    for m, v in list(magics.items()):
+def __by_version(magic_versions: Dict[bytes, str]) -> dict:
+    for m, version in list(magic_versions.items()):
         if m not in by_magic:
-            by_magic[m] = {v}
+            by_magic[m] = {version}
         else:
-            by_magic[m].add(v)
-        by_version[v] = m
+            by_magic[m].add(version)
+        by_version[version] = m
     return by_version
 
 
@@ -111,7 +112,7 @@ PYTHON_MAGIC_INT = magic2int(MAGIC)
 # Apple MPW compiler swaps their values, botching string constants.
 #
 # The magic numbers must be spaced apart at least 2 values, as the
-# -U interpeter flag will cause MAGIC+1 being used. They have been
+# -U interpreter flag will cause MAGIC+1 being used. They have been
 # odd numbers for some time now.
 #
 # There were a variety of old schemes for setting the magic number.
@@ -127,8 +128,10 @@ PYTHON_MAGIC_INT = magic2int(MAGIC)
 # that's come before.  The tags are included in the following table, starting
 # with Python 3.2a0.
 
-# The below is taken from from Python/import.c, and more recently
+# The below is taken from Python/import.c, and more recently
 # Lib/importlib/_bootstrap.py and other sources
+# And more recently:
+# https://github.com/google/pytype/blob/main/pytype/pyc/magic.py
 
 #                  magic,  canonic version number
 add_magic_from_int(39170, "1.0")
@@ -162,7 +165,7 @@ add_magic_from_int(62111, "2.5b3")  # fix wrong code: x += yield
 # have been removed
 add_magic_from_int(62121, "2.5c1")
 
-# Fix wrong code: "for x, in ..." in listcomp/genexp)
+# Fix wrong code: "for x, in ..." in listcomp/genexp
 add_magic_from_int(62131, "2.5c2")
 
 # Dropbox-modified Python 2.5 used in versions 1.1x and before of Dropbox
@@ -184,7 +187,7 @@ add_magic_from_int(62211, "2.7")  # introduce MAP_ADD and SET_ADD
 
 add_magic_from_int(2657, "2.7pyston-0.6.1")
 
-# PyPy including pypy-2.6.1, pypy-5.0.1 PyPy adds 7 to the corresponding CPython nmber
+# PyPy including pypy-2.6.1, pypy-5.0.1 PyPy adds 7 to the corresponding CPython number
 add_magic_from_int(62211 + 7, "2.7pypy")
 
 add_magic_from_int(3000, "3.000")
@@ -201,7 +204,7 @@ add_magic_from_int(3091, "3.000+10")  # kill str8 interning
 add_magic_from_int(3101, "3.000+11")  # merge from 2.6a0, see 62151
 add_magic_from_int(3103, "3.000+12")  # __file__ points to source file
 add_magic_from_int(3111, "3.0a4")  # WITH_CLEANUP optimization
-add_magic_from_int(3131, "3.0a5")  # lexical exception stacking, including POP_EXCEPT)
+add_magic_from_int(3131, "3.0a5")  # lexical exception stacking, including POP_EXCEPT
 add_magic_from_int(3141, "3.1a0")  # optimize list, set and dict comprehensions
 add_magic_from_int(3151, "3.1a0+")  # optimize conditional branches
 add_magic_from_int(3160, "3.2a0")  # add SETUP_WITH
@@ -224,7 +227,7 @@ add_magic_from_int(3220, "3.3a1")  # changed PEP 380 implementation
 add_magic_from_int(3210, "3.3a2")
 add_magic_from_int(3230, "3.3a4")  # revert changes to implicit __class__ closure
 
-# Evaluate positional default arg keyword-only defaults)
+# Evaluate positional default arg keyword-only defaults
 add_magic_from_int(3250, "3.4a1")
 
 # Add LOAD_CLASSDEREF; add_magic_from_int locals, f class to override free vars
@@ -255,14 +258,14 @@ add_magic_from_int(
 )  # add SETUP_ANNOTATIONS and STORE_ANNOTATION opcodes #27985
 add_magic_from_int(
     3376, "3.6b1+2"
-)  # simplify CALL_FUNCTIONs & BUILD_MAP_UNPACK_WITH_CALL
+)  # simplify CALL_FUNCTION* & BUILD_MAP_UNPACK_WITH_CALL
 add_magic_from_int(3377, "3.6b1+3")  # set __class__ cell from type.__new__ #23722
 add_magic_from_int(3378, "3.6b2")  # add BUILD_TUPLE_UNPACK_WITH_CALL #28257
 
 # more thorough __class__ validation #23722
 add_magic_from_int(3379, "3.6rc1")
 
-# add LOAD_METHOD and CALL_METHOD opcodes #26110)
+# add LOAD_METHOD and CALL_METHOD opcodes #26110
 add_magic_from_int(3390, "3.7.0alpha0")
 
 # update GET_AITER #31709
@@ -277,7 +280,7 @@ add_magic_from_int(3392, "3.7.0beta2")
 add_magic_from_int(3393, "3.7.0beta3")
 
 # restored docstring as the first stmt in the body; this might
-# affected the first line number #32911)
+# affect the first line number #32911
 add_magic_from_int(3394, "3.7.0")
 
 # move frame block handling to compiler #17611
@@ -294,7 +297,7 @@ add_magic_from_int(3410, "3.8.0a1+")
 add_magic_from_int(3411, "3.8.0b2+")
 
 # Swap the position of positional args and positional only args in
-# ast.arguments #37593)
+# ast.arguments #37593
 add_magic_from_int(3412, "3.8.0beta2")
 
 # Fix "break" and "continue" in "finally" #37830
@@ -329,8 +332,59 @@ add_magic_from_int(3432, "3.10a2")
 
 # RERAISE restores f_lasti if oparg != 0
 add_magic_from_int(3433, "3.10a2")
+add_magic_from_int(3434, "3.10a6")
+add_magic_from_int(3435, "3.10a7")
+add_magic_from_int(3438, "3.10b1")
 add_magic_from_int(3439, "3.10.0rc2")
 
+add_magic_from_int(3450, "3.11a1a")
+add_magic_from_int(3451, "3.11a1b")
+add_magic_from_int(3452, "3.11a1c")
+add_magic_from_int(3453, "3.11a1d")
+add_magic_from_int(3454, "3.11a1e")
+add_magic_from_int(3455, "3.11a1f")
+add_magic_from_int(3457, "3.11a1g")
+add_magic_from_int(3458, "3.11a1h")
+add_magic_from_int(3459, "3.11a1i")
+add_magic_from_int(3460, "3.11a1j")
+add_magic_from_int(3461, "3.11a1k")
+add_magic_from_int(3462, "3.11a2")
+add_magic_from_int(3463, "3.11a3a")
+add_magic_from_int(3464, "3.11a3b")
+add_magic_from_int(3465, "3.11a4a")
+add_magic_from_int(3466, "3.11a4b")
+add_magic_from_int(3466, "3.11a4c")
+add_magic_from_int(3467, "3.11a4d")
+add_magic_from_int(3468, "3.11a4e")
+add_magic_from_int(3469, "3.11a4f")
+add_magic_from_int(3470, "3.11a4g")
+add_magic_from_int(3471, "3.11a4h")
+add_magic_from_int(3472, "3.11a4i")
+add_magic_from_int(3473, "3.11a4j")
+add_magic_from_int(3474, "3.11a4k")
+add_magic_from_int(3475, "3.11a5a")
+add_magic_from_int(3476, "3.11a5b")
+add_magic_from_int(3477, "3.11a5c")
+add_magic_from_int(3478, "3.11a5d")
+add_magic_from_int(3479, "3.11a5e")
+add_magic_from_int(3480, "3.11a5e")
+add_magic_from_int(3481, "3.11a5f")
+add_magic_from_int(3482, "3.11a5g")
+add_magic_from_int(3483, "3.11a5h")
+add_magic_from_int(3484, "3.11a5i")
+add_magic_from_int(3485, "3.11a5j")
+add_magic_from_int(3486, "3.11a6a")
+add_magic_from_int(3487, "3.11a6b")
+add_magic_from_int(3488, "3.11a6c")
+add_magic_from_int(3489, "3.11a6d")
+add_magic_from_int(3490, "3.11a6d")
+add_magic_from_int(3491, "3.11a7a")
+add_magic_from_int(3492, "3.11a7b")
+add_magic_from_int(3493, "3.11a7c")
+add_magic_from_int(3494, "3.11a7d")
+add_magic_from_int(3495, "3.11a7e")
+add_magic_from_int(3531, "3.12.0rc2")
+
 # Weird ones
 # WTF? Python 3.2.5 and PyPy have weird magic numbers
 
@@ -342,34 +396,33 @@ add_magic_from_int(192, "3.6pypy")  # '3.6.9 ... PyPy 7.1.0-beta0'
 add_magic_from_int(224, "3.7pypy")  # PyPy 3.7.9-beta0
 add_magic_from_int(240, "3.7pypy")  # PyPy 3.7.9-beta0
 add_magic_from_int(256, "3.8pypy")  # PyPy 3.8.15
-add_magic_from_int(336, "3.9pypy")  # PyPy 3.9.15
+add_magic_from_int(336, "3.9pypy")  # PyPy 3.9.15, PyPy 3.9.17
+add_magic_from_int(384, "3.10pypy")  # PyPy 3.10.12
 
 # NOTE: This is JVM bytecode not Python bytecode
 add_magic_from_int(21150, "3.8.5Graal")
 
+# Graal Java 21.0.2
+add_magic_from_int(21280, "3.10.8Graal")
+
+
 add_magic_from_int(1011, "2.7.1b3Jython")  # jython
 add_magic_from_int(22138, "2.7.7Pyston")  # 2.7.8pyston, pyston-0.6.0
 
 magics = __by_version(versions)
 magics["3.8.12pypy"] = magics["3.8.0rc1+"]
 magics["3.9.15pypy"] = magics["3.9.0alpha1"]
+magics["3.9.16pypy"] = magics["3.9.0alpha1"]
 
 # From a Python version given in sys.info, e.g. 3.6.1,
 # what is the "canonic" version number, e.g. '3.6.0rc1'
 canonic_python_version = {}
 
 
-def add_canonic_versions(versions, canonic):
-    for v in versions.split():
-        canonic_python_version[v] = canonic
-        magics[v] = magics[canonic]
-        try:
-            magics[float(v)] = magics[canonic]
-        except Exception:
-            pass
-        pass
-
-    return
+def add_canonic_versions(release_versions: str, canonic):
+    for version in release_versions.split():
+        canonic_python_version[version] = canonic
+        magics[version] = magics[canonic]
 
 
 add_canonic_versions("1.5.1 1.5.2", "1.5")
@@ -410,40 +463,57 @@ add_canonic_versions(
 add_canonic_versions("3.7b1", "3.7.0beta3")
 add_canonic_versions("3.8a1", "3.8.0beta2")
 
-add_canonic_versions("2.7.10pypy 2.7.12pypy 2.7.13pypy", "2.7pypy")
+add_canonic_versions("2.7.10pypy 2.7.12pypy 2.7.13pypy 2.7.18pypy", "2.7pypy")
 add_canonic_versions("2.7.3b0Jython", "2.7.1b3Jython")
 add_canonic_versions("3.8.5Graal", "3.8.5Graal")
+add_canonic_versions("3.8.10Graal", "3.8.0rc1+")
 
 add_canonic_versions("3.2.5pypy", "3.2pypy")
 add_canonic_versions("3.3.5pypy", "3.3pypy")
 add_canonic_versions("3.5.3pypy", "3.5pypy")
 add_canonic_versions("3.6.9pypy", "3.6pypy")
 add_canonic_versions("3.7.0pypy 3.7.9pypy 3.7.10pypy 3.7.12pypy 3.7.13pypy", "3.7pypy")
-add_canonic_versions("3.8.0pypy 3.8pypy 3.8.12pypy 3.8.13pypy 3.8.15pypy", "3.8.12pypy")
+add_canonic_versions("3.8.0pypy 3.8.12pypy 3.8.13pypy 3.8.15pypy", "3.8.12pypy")
+add_canonic_versions("3.8.16pypy", "3.8pypy")
+add_canonic_versions("3.9.17pypy 3.9.18pypy", "3.9pypy")
+add_canonic_versions("3.10.12pypy 3.10.13pypy 3.10pypy", "3.10pypy")
 add_canonic_versions("2.7.8Pyston", "2.7.7Pyston")
 add_canonic_versions("3.7.0alpha3", "3.7.0alpha3")
 add_canonic_versions(
     "3.7 3.7.0beta5 3.7.1 3.7.2 3.7.3 3.7.4 3.7.5 3.7.6 3.7.7 3.7.8 3.7.9 "
-    "3.7.10 3.7.11 3.7.12 3.7.13 3.7.14 3.7.15 3.7.16",
+    "3.7.10 3.7.11 3.7.12 3.7.13 3.7.14 3.7.15 3.7.16 3.7.17",
     "3.7.0",
 )
 add_canonic_versions("3.8.0alpha0 3.8.0alpha3 3.8.0a0", "3.8.0a3+")
 add_canonic_versions(
-    "3.8b4 3.8.0candidate1 3.8 3.8.0 3.8.1 3.8.2 3.8.3 3.8.4 3.8.5 3.8.6 3.8.7 3.8.8 3.8.9 "
-    "3.8.10 3.8.11 3.8.12 3.8.13 3.8.14 3.8.15 3.8.16",
+    "3.8b4 3.8.0candidate1 3.8 3.8.0 3.8.1 3.8.2 3.8.3 3.8.4 3.8.5 3.8.6 3.8.7 3.8.8 "
+    "3.8.9 3.8.10 3.8.11 3.8.12 3.8.13 3.8.14 3.8.15 3.8.16 3.8.17 3.8.18 ",
     "3.8.0rc1+",
 )
 add_canonic_versions(
     "3.9 3.9.0 3.9.0a1+ 3.9.0a2+ 3.9.0alpha1 3.9.0alpha2", "3.9.0alpha1"
 )
 add_canonic_versions(
-    "3.9 3.9.0 3.9.1 3.9.2 3.9.3 3.9.4 3.9.5 3.9.6 3.9.7 3.9.8 3.9.9 3.9.10 3.9.11 3.9.12 3.9.13 3.9.14 "
-    "3.9.14 3.9.15 3.9.16 3.9.10pypy 3.9.11pypy 3.9.12pypy 3.9.15pypy 3.9.0b5+",
+    "3.9 3.9.0 3.9.1 3.9.2 3.9.3 3.9.4 3.9.5 3.9.6 3.9.7 3.9.8 3.9.9 3.9.10 3.9.11 "
+    "3.9.12 3.9.13 3.9.14 3.9.14 3.9.15 3.9.16 3.9.10pypy 3.9.11pypy 3.9.12pypy "
+    "3.9.15pypy 3.9.16pypy 3.9.0b5+ 3.9.17 3.9.18",
     "3.9.0beta5",
 )
 
 add_canonic_versions(
-    "3.10 3.10.0 3.10.1 3.10.2 3.10.3 3.10.4 3.10.5 3.10.6 3.10.7 3.10.8 3.10.9", "3.10.0rc2"
+    "3.10 3.10.0 3.10.1 3.10.2 3.10.3 3.10.4 3.10.5 3.10.6 3.10.7 3.10.8 3.10.9 "
+    "3.10.10 3.10.11 3.10.12 3.10.13",
+    "3.10.0rc2",
+)
+
+add_canonic_versions(
+    "3.11 3.11.0 3.11.1 3.11.2 3.11.3 3.11.4 3.11.5 3.11.6 3.11.7 3.11.8",
+    "3.11a7e",
+)
+
+add_canonic_versions(
+    "3.12 3.12.0 3.12.1 3.12.2",
+    "3.12.0rc2",
 )
 
 # The canonic version for a canonic version is itself
@@ -461,8 +531,8 @@ def magic_int2tuple(magic_int: int) -> tuple:
     """Convert a Python magic int into a 'canonic' tuple
     e.g. (2, 7), (3, 7). runtime error is raised if "version" is not found.
 
-    Note that there can be several magic_int's that map to a single floating-
-    point number. For example 3320 (3.5.a0), 3340 (3.5b1)
+    Note that there can be several magic_int's that map to a single floating-point
+    number. For example 3320 (3.5.a0), 3340 (3.5b1)
     all map to 3.5.
     """
     return py_str2tuple(magicint2version[magic_int])
@@ -482,18 +552,17 @@ def py_str2tuple(orig_version):
     if version in magics:
         m = re.match(r"^(\d)\.(\d+)\.(\d+)", version)
         if m:
-            return (int(m.group(1)), int(m.group(2)), int(m.group(3)))
+            return int(m.group(1)), int(m.group(2)), int(m.group(3))
         else:
             # Match things like 3.5a0, 3.5b2, 3.6a1+1, 3.6rc1, 3.7.0beta3
-            m = re.match(r"^(\d)\.(\d)(\d+)?[abr]?", version)
+            m = re.match(r"^(\d)\.(\d(\d+)?)[abr]?", version)
             if m:
-                return (int(m.group(1)), int(m.group(2)))
+                return int(m.group(1)), int(m.group(2))
             pass
         pass
     raise RuntimeError(
         "Can't find a valid Python version for version %s" % orig_version
     )
-    return
 
 
 def sysinfo2magic(version_info=sys.version_info) -> bytes:
@@ -508,12 +577,14 @@ def sysinfo2magic(version_info=sys.version_info) -> bytes:
 
     if IS_PYPY:
         vers_str += "pypy"
+    elif IS_GRAAL:
+        vers_str += "Graal"
     else:
         try:
             import platform
 
             platform_str = platform.python_implementation()
-            if platform in ("Jython", "Pyston", "GraalVM"):
+            if platform_str in ("Jython", "Pyston", "GraalVM"):
                 vers_str += platform_str
                 pass
         except ImportError:
@@ -532,7 +603,7 @@ def test():
     print("This Python interpreter has versions:", magic_current)
     print("Magic code: ", PYTHON_MAGIC_INT)
     print(type(magic_20), len(magic_20), repr(magic_20))
-    assert sysinfo2magic() == MAGIC
+    assert sysinfo2magic() == MAGIC, (sysinfo2magic(), MAGIC)
 
 
 if __name__ == "__main__":
